
package org.gvsig.remoteclient.wms;

import java.util.Vector;

import es.prodevelop.gvsig.mini.geom.Extent;



/**
 * Describes the status of a WMSclient, so it adds to the Remote client status
 * a list of layers, a list of layer styles, the extent of the map.
 * Provides the functionality to modify these lists.
 * 
 */
public class WMSStatus extends org.gvsig.remoteclient.RemoteClientStatus {
			
	// list of layer to be retrieved by the WMS
    private Vector layers;
    // List of layer styles
    private Vector styles;
    private Vector dimensions;
    // extent required by the WMS client
    private Extent extent;
    private boolean transparency;
	private String onlineResource;
	private String infoFormat = "";
    
	public WMSStatus()
    {
    	layers = new Vector();
    	styles = new Vector();    	    	
    }
    
    // sets the list of layers required by the WMS client
    public void setLayerNames(Vector _layers) {        
        layers = _layers;
    } 

/**
 * <p>Retrieves the layer list required by the WMS client</p>
 * 
 * 
 * @return Vector, the list of layers
 */
    public Vector getLayerNames() {        
        return layers;
    } 

/**
 * <p> Adds a layer to the list of layers required by the WMS client</p>
 * 
 * 
 * @param _layerName, name of the layer to be added. 
 */
    public void addLayerName(String _layerName) {        
        layers.add(_layerName);
    } 

/**
 * <p>removes a layer from the layer list</p>
 * 
 * 
 * @param _layerName 
 * @return true if the layer name has been deleted from the list
 */
    public boolean removeLayerName(String _layerName) {
    	return layers.remove(_layerName);       
    } 

/**
 * <p>gets the styles list required by the WMS client</p>
 * 
 * 
 * @return Vector with the list of layer styles
 */
    public Vector getStyles() {        
        return styles;
    } 

/**
 * <p>sets the styles list required by the WMS client</p>
 * 
 * 
 * @param _styles, list to be set as the required styles.
 */
    public void setStyles(Vector _styles) {        
        styles = _styles;
    }
    
    /**
     * <p>sets the styles list required by the WMS client</p>
     * 
     * 
     * @param _styles, list to be set as the required styles.
     */
    public void setDimensions(Vector _dimensions) {        
        
        dimensions = _dimensions;
    } 

/**
 * <p>Adds a style name to the styles list required by the WMS client</p>
 * 
 * 
 * @param _name, style name to be added
 */
    public void addStyleName(String _name) {        
        styles.add( _name);
    } 

/**
 * <p>Removes a style from the list of styles required by the WMS client</p>
 * 
 * 
 * @param _name, style name to be removed
 */
    public boolean removeStyleName(String _name) {        
        return styles.remove(_name);
    } 
/**
 * <p>Gets the extent defined by the map</p>
 */
    public Extent getExtent() {        
        return extent;
    } 
/**
 * <p>Sets the extent defined by the map</p>
 */
    public void setExtent(Extent extent) {        
        this.extent = extent;
    } 
 
   

    /**
     * @return
     */
    public boolean getTransparency() {
        return transparency;
    }

    /**
     * @param wmsTransparency
     */
    public void setTransparency(boolean wmsTransparency) {
        transparency = wmsTransparency;
    }

    /**
     * @return
     */
    public Vector getDimensions() {
        return dimensions;
    }    

    /**
     * Returns the URL that the server specified for a WMS request if any was described in
     * its capabilities document. 
     * @param operationName, a String containing the name of the operation (case-independent)
     * @return <b>String</b> containing the URL for this operationName or <B>null</B> if none was
     *                specified.
     */
	public String getOnlineResource() {
		return onlineResource;
	}
	
	/**
	 * Sets the string literal containing the URL of an online resource for a specific
	 * WMS request.
	 * @param operationName, String telling to which request correspond the address
	 * @param url, String containing the URL for the given WMS request
	 */
	public void setOnlineResource(String url) {
		onlineResource = url;
	}

	public void setInfoFormat(String infoFormat) {
		this.infoFormat = infoFormat;
	}

	public String getInfoFormat() {
		return infoFormat;
	}

}
