package org.gvsig.remoteclient.wms.util;

import bm.core.tools.StringTokenizer;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.geom.Point;

public class Rectangle2D {

	private double minX;
	private double minY;
	private double maxX;
	private double maxY;
	private Point center = null;

	/**
	 * Constructor. It also sets the center according to the parameters.
	 * 
	 * @param minX
	 *            The minimum x value
	 * @param minY
	 *            The minimum y value
	 * @param maxX
	 *            The maximum x value
	 * @param maxY
	 *            The maximum y value
	 */
	public Rectangle2D(final double minX, final double minY, final double maxX,
			final double maxY) {
		this.minX = minX;
		this.minY = minY;
		this.maxX = maxX;
		this.maxY = maxY;

		center = new Point(0, 0);

		this.calculateCenter();
	}

	/**
	 * Constructor. It also sets the center according to the parameters
	 * 
	 * @param leftBottom
	 *            A <code>Point</code> with the minimum x and y values
	 * @param rightTop
	 *            A <code>Point</code> with the maximum x and y values
	 */
	public Rectangle2D(final Point leftBottom, final Point rightTop) {
		this.minX = leftBottom.getX();
		this.minY = leftBottom.getY();
		this.maxX = rightTop.getX();
		this.maxY = rightTop.getY();

		center = new Point(0, 0);

		this.calculateCenter();
	}

	/**
	 * sets to NULL the center Point
	 */
	public void destroy() {
		this.center = null;
	}

	/**
	 * Clones this instance
	 * 
	 * @return a cloned instance of this Rectangle2D
	 */
	public Object clone() {
		return new Rectangle2D(this.minX, this.minY, this.maxX, this.maxY);
	}

	/**
	 * Compares this Rectangle2D whith other
	 * 
	 * @param Rectangle2D
	 *            The Rectangle2D to compare
	 * @return true if both are equals
	 */
	public boolean equals(final Rectangle2D Rectangle2D) {
		return (this.minX == Rectangle2D.minX && this.maxX == Rectangle2D.maxX
				&& this.minY == Rectangle2D.minY && this.maxY == Rectangle2D.maxY);
	}

	/**
	 * Checks if the Rectangle2D contains the x and y values
	 * 
	 * @param x
	 *            X value
	 * @param y
	 *            Y value
	 * @return true if contains
	 */
	public boolean contains(final double x, final double y) {
		return ((x >= this.minX) && (x <= this.maxX) && (y >= this.minY) && (y <= this.maxY));
	}

	/**
	 * Checks if the Rectangle2D contains the x and y values
	 * 
	 * @param point
	 *            an array with x and y values
	 * @return true if contains
	 */
	public boolean contains(final double[] point) {
		return ((point[0] >= this.minX) && (point[0] <= this.maxX)
				&& (point[1] >= this.minY) && (point[1] <= this.maxY));
	}

	/**
	 * Check if the Rectangle2Ds contains the Point
	 * 
	 * @param point
	 *            The point to check
	 * @return true if contains
	 */
	public boolean contains(final Point point) {
		return this.contains(point.getX(), point.getY());
	}

	/**
	 * Checks if the Rectangle2D intersects with other
	 * 
	 * @param Rectangle2D
	 *            The Rectangle2D to check
	 * @return true if intersects
	 */
	public boolean intersect(final Rectangle2D Rectangle2D) {

		final boolean inBottom = (Rectangle2D.minY == this.minY && Rectangle2D.maxY == this.maxY) ? true
				: (((Rectangle2D.minY > this.minY) && (Rectangle2D.minY < this.maxY)) || ((this.minY > Rectangle2D.minY) && (this.minY < Rectangle2D.maxY)));
		final boolean inTop = (Rectangle2D.minY == this.minY && Rectangle2D.maxY == this.maxY) ? true
				: (((Rectangle2D.maxY > this.minY) && (Rectangle2D.maxY < this.maxY)) || ((this.maxY > Rectangle2D.minY) && (this.maxY < Rectangle2D.maxY)));
		final boolean inRight = (Rectangle2D.maxX == this.maxX && Rectangle2D.minX == this.minX) ? true
				: (((Rectangle2D.maxX > this.minX) && (Rectangle2D.maxX < this.maxX)) || ((this.maxX > Rectangle2D.minX) && (this.maxX < Rectangle2D.maxX)));
		final boolean inLeft = (Rectangle2D.maxX == this.maxX && Rectangle2D.minX == this.minX) ? true
				: (((Rectangle2D.minX > this.minX) && (Rectangle2D.minX < this.maxX)) || ((this.minX > Rectangle2D.minX) && (this.minX < Rectangle2D.maxX)));

		return (this.contains(Rectangle2D) || Rectangle2D.contains(this) || ((inTop || inBottom) && (inLeft || inRight)));
	}

	/**
	 * Check if the Rectangle2D intersects and returns an Rectangle2D that
	 * represent the area that intersects
	 * 
	 * @param Rectangle2D
	 *            The Rectangle2D to check
	 * @param ext
	 *            true or false it doesn't matter
	 * @return The intersection Rectangle2D. NULL if the Rectangle2Ds don't
	 *         intersect
	 */
	public Rectangle2D intersect(final Rectangle2D Rectangle2D,
			final boolean ext) {
		if (this.intersect(Rectangle2D)) {
			return new Rectangle2D(Math.max(Rectangle2D.getMinX(), this
					.getMinX()), Math
					.max(Rectangle2D.getMinY(), this.getMinY()), Math.min(
					Rectangle2D.getMaxX(), this.getMaxX()), Math.min(
					Rectangle2D.getMaxY(), this.getMaxY()));
		}
		return null;
	}

	/**
	 * Checks if the Rectangle2D contains the other
	 * 
	 * @param Rectangle2D
	 *            The Rectangle2D to check
	 * @return true if contains
	 */
	public boolean contains(final Rectangle2D Rectangle2D) {

		final boolean inLeft = (Rectangle2D.minX >= this.minX)
				&& (Rectangle2D.minX <= this.maxX);
		final boolean inTop = (Rectangle2D.maxY >= this.minY)
				&& (Rectangle2D.maxY <= this.maxY);
		final boolean inRight = (Rectangle2D.maxX >= this.minX)
				&& (Rectangle2D.maxX <= this.maxX);
		final boolean inBottom = (Rectangle2D.minY >= this.minY)
				&& (Rectangle2D.minY <= this.maxY);

		return (inTop && inLeft && inBottom && inRight);
	}

	/**
	 * Calculates the area of the Rectangle2D
	 * 
	 * @return The area
	 */
	public double area() {
		return this.getWidth() * this.getHeight();
	}

	private void calculateCenter() {
		this.center.setX((minX + maxX) / 2);
		this.center.setY((minY + maxY) / 2);
	}

	/**
	 * 
	 * @return An instance of a Point with the minimum x and y values
	 */
	public Point getLefBottomCoordinate() {
		return new Point(this.minX, this.minY);
	}

	/**
	 * 
	 * @return @return An instance of a Point with the maximum x and y values
	 */
	public Point getRightTopCoordinate() {
		return new Point(this.maxX, this.maxY);
	}

	/**
	 * Sets the minimum x and y values
	 * 
	 * @param leftBottom
	 *            A Point with the x and y values to set
	 */
	public void setLeftBottomCoordinate(final Point leftBottom) {
		this.minX = leftBottom.getX();
		this.minY = leftBottom.getY();
	}

	/**
	 * Sets the maximum x and y values
	 * 
	 * @param rightTop
	 *            A Point with the x and y values to set
	 */
	public void setRightTopCoordinate(final Point rightTop) {
		this.maxX = rightTop.getX();
		this.maxY = rightTop.getY();
	}

	/**
	 * This method is used to get a printable String with the x,y values of the
	 * Rectangle2D
	 * 
	 * @return a String with the minimum and maximum x,y values of the
	 *         Rectangle2D in this format:
	 *         "minX= 0.0, minY= 0.0, maxX= 1.0, maxY= 1.0"
	 */
	public String toPrintString() {
		return new StringBuffer().append("minX= ").append(this.minX).append(
				", minY= ").append(this.minY).append(", maxX= ").append(
				this.maxX).append(", maxY= ").append(this.maxY).toString();
	}

	/**
	 * This methos is used to build the URL of a tile request
	 * 
	 * @return a String with the minimum and maximum x,y values of the
	 *         Rectangle2D in this format: "0.0, 0.0, 1.0, 1.0"
	 */
	public String toString() {
		return new StringBuffer().append(this.minX).append(",").append(
				this.minY).append(",").append(this.maxX).append(",").append(
				this.maxY).toString();
	}

	/**
	 * This methos is used to build the URL of a tile request
	 * 
	 * @return a StringBuffer with the minimum and maximum x,y values of the
	 *         Rectangle2D in this format: "0.0, 0.0, 1.0, 1.0"
	 */
	public StringBuffer toStringBuffer() {
		StringBuffer s = new StringBuffer();
		s.append(this.minX).append(",").append(this.minY).append(",").append(
				this.maxX).append(",").append(this.maxY);
		return s;
	}

	/**
	 * Calculates the width of the Rectangle2D
	 * 
	 * @return The width of the Rectangle2D
	 */
	public double getWidth() {
		return Math.abs(maxX - minX);
	}

	/**
	 * Calculates the height of the Rectangle2D
	 * 
	 * @return The height of the Rectangle2D
	 */
	public double getHeight() {
		return Math.abs(maxY - minY);
	}

	/**
	 * @return the minX
	 */
	public double getMinX() {
		return minX;
	}

	/**
	 * @param minX
	 *            the minX to set
	 */
	public void setMinX(final double minX) {
		this.minX = minX;
	}

	/**
	 * @return the minY
	 */
	public double getMinY() {
		return minY;
	}

	/**
	 * @param minY
	 *            the minY to set
	 */
	public void setMinY(final double minY) {
		this.minY = minY;
	}

	/**
	 * @return the maxX
	 */
	public double getMaxX() {
		return maxX;
	}

	/**
	 * @param maxX
	 *            the maxX to set
	 */
	public void setMaxX(final double maxX) {
		this.maxX = maxX;
	}

	/**
	 * @return the maxY
	 */
	public double getMaxY() {
		return maxY;
	}

	/**
	 * @param maxY
	 *            the maxY to set
	 */
	public void setMaxY(final double maxY) {
		this.maxY = maxY;
	}

	/**
	 * @return the center
	 */
	public Point getCenter() {
		this.calculateCenter();
		return center;
	}

	/**
	 * @param center
	 *            the center to set
	 */
	public void setCenter(final Point center) {
		this.center = center;
	}

	/**
	 * parses a String into a Rectangle2D. The String should have the same
	 * format as the method toString() returns.
	 * 
	 * @param Rectangle2D
	 * @return
	 */
	public static Rectangle2D parseString(final String Rectangle2D) {
		StringTokenizer s = null;
		try {
			s = new StringTokenizer(Rectangle2D, ",");
			if (s.length < 4)
				return null;
			return new Rectangle2D(Double.parseDouble(s.nextToken()), Double
					.parseDouble(s.nextToken()), Double.parseDouble(s
					.nextToken()), Double.parseDouble(s.nextToken()));
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		} finally {
			s = null;
		}
	}
}
