
package org.gvsig.remoteclient.wms.wms_1_1_0;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.TreeMap;

import org.gvsig.remoteclient.utils.CapabilitiesTags;
import org.gvsig.remoteclient.utils.EncodingXMLParser;
import org.gvsig.remoteclient.utils.ExceptionTags;
import org.gvsig.remoteclient.wms.WMSServiceInformation;
import org.gvsig.remoteclient.wms.WMSStatus;
import org.gvsig.remoteclient.wms.request.WMSGetCapabilitiesRequest;
import org.gvsig.remoteclient.wms.request.WMSGetFeatureInfoRequest;
import org.gvsig.remoteclient.wms.request.WMSGetLegendGraphicRequest;
import org.gvsig.remoteclient.wms.request.WMSGetMapRequest;
import org.gvsig.remoteclient.wms.wms_1_1_0.request.WMSGetCapabilitiesRequest1_1_0;
import org.gvsig.remoteclient.wms.wms_1_1_0.request.WMSGetFeatureInfoRequest1_1_0;
import org.gvsig.remoteclient.wms.wms_1_1_0.request.WMSGetLegendGraphicRequest1_1_0;
import org.gvsig.remoteclient.wms.wms_1_1_0.request.WMSGetMapRequest1_1_0;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

/**
 * <p>
 * Describes the handler to comunicate to a WMS 1.1.0
 * </p>
 */
public class WMSProtocolHandler1_1_0 extends org.gvsig.remoteclient.wms.WMSProtocolHandler
{
	private WMSLayer1_1_0 fakeRootLayer;
    
	public WMSProtocolHandler1_1_0()
	{
		this.version = "1.1.0";
		this.name = "WMS1.1.0";
		this.serviceInfo = new WMSServiceInformation(); 
		this.layers = new TreeMap();
	}
    
//------------------------------------------------------------------------------
// Parsing methods....    
//------------------------------------------------------------------------------    


	public boolean parseCapabilities(File f)
    {       
    	rootLayer = null;
    	FileReader reader = null;
		try {
			reader = new FileReader(f);

			//	    		FileInputStream fis = new FileInputStream(f);
			//	    		InputStreamReader reader = new InputStreamReader(fis,"ISO-8859-1");
			// BufferedReader br = new BufferedReader(reader);

			char[] buffer = new char[100];
			reader.read(buffer);
			StringBuffer strBuff = new StringBuffer();
			strBuff.append(buffer);
			String auxStr = strBuff.toString().toLowerCase();
			int a = auxStr.toLowerCase().indexOf("<?xml");
			if (a != -1)
				auxStr = auxStr.substring(a, auxStr.length());

			StringBuffer st = new StringBuffer(auxStr);
			String searchText = "encoding=\"";
			int index = st.indexOf(searchText);

			if (index > -1) {
				st.delete(0, index + searchText.length());
				encoding = st.substring(0, st.indexOf("\""));
			}
			
			reader.close();
			
		} catch (Exception ex) {
//			logger.error("In main parse method: " + ex.getMessage());
		}

		int tag;
		KXmlParser kxmlParser = null;
		kxmlParser = new KXmlParser();
		try {
//			reader = new FileReader(f);
//			//	    		FileInputStream fis = new FileInputStream(f);
//			//	    		InputStreamReader reader = new InputStreamReader(fis,"ISO-8859-1");
//			// BufferedReader br = new BufferedReader(reader);
//			// patch for ArcIMS + WMS connector > 9.0 bug
//			
//			char[] buffer = new char[(int) f.length()];
//			
//			logger.debug("PHONEME - buffer.len = " + buffer.length);
//			
//			reader.read(buffer);
//			
//			StringBuffer strBuff = new StringBuffer();
//			strBuff.append(buffer);
//			String auxStr = strBuff.toString();
//			
//			logger.debug("PHONEME - auxStr = " + auxStr);
//			// String string = new String(buffer);
//			
//			int a = auxStr.toLowerCase().indexOf("<?xml");
//			
//			logger.debug("PHONEME - a = " + a);
//			
//			if (a != -1) {
//				auxStr = auxStr.substring(a, auxStr.length());
//				logger.debug("PHONEME - auxStr = " + auxStr);
//				kxmlParser.setInput(new StringReader(auxStr));
//			} else
//				// end patch
//				kxmlParser.setInput(new FileInputStream(f), encoding);

			
			FileInputStream fir = new FileInputStream(f);
			kxmlParser.setInput(fir, encoding);
    		kxmlParser.nextTag();
    		if ( kxmlParser.getEventType() != KXmlParser.END_DOCUMENT ) 
    		{    		
    			kxmlParser.require(KXmlParser.START_TAG, null, CapabilitiesTags.CAPABILITIES_ROOT1_1_1);    			
    			tag = kxmlParser.nextTag();
				 while(tag != KXmlParser.END_DOCUMENT)
				 {
                     switch(tag)
					 {
                         
						case KXmlParser.START_TAG:
							if (kxmlParser.getName().compareTo(CapabilitiesTags.SERVICE )==0)
							{
								parseServiceTag(kxmlParser);
							}	
							else if (kxmlParser.getName().compareTo(CapabilitiesTags.CAPABILITY)==0)
							{
								parseCapabilityTag(kxmlParser);
							}
							break;
						case KXmlParser.END_TAG:							
							break;
						case KXmlParser.TEXT:
												
						break;
					 }
    				 tag = kxmlParser.next();
    			 }

    			kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
    		}
    	}
    	catch(XmlPullParserException parser_ex){
    		parser_ex.printStackTrace();
    		return false;
    	}
   		catch (IOException ioe) {			
   			ioe.printStackTrace();
   			return false;
		} finally {
            return true;
        }
    } 
    
    /**
     * <p>Parses the Service Information </p>
     */    
    private void parseServiceTag(KXmlParser parser) throws IOException, XmlPullParserException 
    {
    	int currentTag;
    	boolean end = false;
    	
    	parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.SERVICE);
    	currentTag = parser.next();
    	
    	while (!end) 
    	{
			 switch(currentTag)
			 {
				case KXmlParser.START_TAG:
					if (parser.getName().compareTo(CapabilitiesTags.NAME)==0)
					{
						serviceInfo.name = parser.nextText(); 
					}	
					else if (parser.getName().compareTo(CapabilitiesTags.TITLE)==0)
					{
						serviceInfo.title = parser.nextText(); 
					}
					else if (parser.getName().compareTo(CapabilitiesTags.ABSTRACT)==0)
					{
						serviceInfo.abstr = parser.nextText(); 
					}
					else if (parser.getName().compareTo(CapabilitiesTags.ONLINERESOURCE)==0)
					{
				    	String value = null;
				        value = parser.getAttributeValue("", CapabilitiesTags.XLINK_HREF);
				        if (value != null){
				        	serviceInfo.online_resource = value;
				        }
					}					
					else if ((parser.getName().compareTo(CapabilitiesTags.KEYWORDLIST)==0) ||
							(parser.getName().compareTo(CapabilitiesTags.CONTACTINFORMATION)==0))
					{
						parser.skipSubTree();
					}					
					break;
				case KXmlParser.END_TAG:
					if (parser.getName().compareTo(CapabilitiesTags.SERVICE) == 0)
						end = true;
					break;
				case KXmlParser.TEXT:					
				break;
			 }
             if (!end)
                 currentTag = parser.next();
    	}
    	parser.require(KXmlParser.END_TAG, null, CapabilitiesTags.SERVICE);
    }
    
    /**
     * <p>Parses the Capability Tag </p>
     */    
    private void parseCapabilityTag(KXmlParser parser) throws IOException, XmlPullParserException
    { 	
    	int currentTag;
    	boolean end = false;
    	
    	parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.CAPABILITY);
    	currentTag = parser.next();
    	
    	while (!end) 
    	{
			 switch(currentTag)
			 {
				case KXmlParser.START_TAG:
					if (parser.getName().compareTo(CapabilitiesTags.REQUEST)==0)
					{
						parseRequestTag(parser); 
					}	
					else if (parser.getName().compareTo(CapabilitiesTags.EXCEPTION)==0)
					{
						// TODO:
						// add the exception formats supported.
					}
					else if (parser.getName().compareTo(CapabilitiesTags.LAYER)==0)
					{
						WMSLayer1_1_0 lyr = new WMSLayer1_1_0();
                        if (rootLayer == null)
                            rootLayer = lyr;
                        else {
                            // Handles when there is no general root layer, will use
                            // a fake non-queryable one.
                            if (!rootLayer.equals(getFakeRootLayer())){
                                WMSLayer1_1_0 aux = (WMSLayer1_1_0) rootLayer;
                                rootLayer  = getFakeRootLayer();
                                rootLayer.getChildren().add(aux);
                            }
                            rootLayer.getChildren().add(lyr);
                        }
						lyr.parse(parser, layers);
						
                        if (lyr.getName()!=null)
						    layers.put(lyr.getName(), lyr); 													
					}
					else if ((parser.getName().compareTo(CapabilitiesTags.VENDORSPECIFICCAPABILITIES)==0) ||
							(parser.getName().compareTo(CapabilitiesTags.USERDEFINEDSYMBOLIZATION )==0))
                            
					{
						parser.skipSubTree();
					}					
					break;
				case KXmlParser.END_TAG:
					if (parser.getName().compareTo(CapabilitiesTags.CAPABILITY) == 0)
						end = true;
					break;
				case KXmlParser.TEXT:					
				break;
			 }
			 if (!end)
			 currentTag = parser.next();
    	}
    	//parser.require(KXmlParser.END_TAG, null, CapabilitiesTags.CAPABILITY);    	
    }   
    
    
    private WMSLayer1_1_0 getFakeRootLayer(){
        if (fakeRootLayer == null){
            fakeRootLayer = new WMSLayer1_1_0();
            fakeRootLayer.setTitle(serviceInfo.title);
            fakeRootLayer.setQueryable(false);
            fakeRootLayer.setName(null);
        }
        return fakeRootLayer;
    }
    
    /* (non-Javadoc)
     * @see org.gvsig.remoteClient.wms.WMSProtocolHandler#parseException(byte[])
     */
    protected String parseException(byte[] data) {
        ArrayList errors = new ArrayList();
        KXmlParser kxmlParser = new KXmlParser();
        
        try
        {
            kxmlParser.setInput(new ByteArrayInputStream(data), encoding);        
            kxmlParser.nextTag();
            int tag;
            if ( kxmlParser.getEventType() != KXmlParser.END_DOCUMENT ) 
            { 
                kxmlParser.require(KXmlParser.START_TAG, null, ExceptionTags.EXCEPTION_ROOT);             
                tag = kxmlParser.nextTag();
                 while(tag != KXmlParser.END_DOCUMENT)
                 {
                     switch(tag)
                     {
                        case KXmlParser.START_TAG:
                            if (kxmlParser.getName().compareTo(ExceptionTags.SERVICE_EXCEPTION)==0){
                                String errorCode = kxmlParser.getAttributeValue("", ExceptionTags.CODE);
                                errorCode = (errorCode != null) ? "["+errorCode+"] " : "";
                                String errorMessage = kxmlParser.nextText();
                                errors.add(errorCode+errorMessage);
                            }
                            break;
                        case KXmlParser.END_TAG:                            
                            break;
                        
                     }
                     tag = kxmlParser.nextTag();
                 }
                 //kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
            }
        }
        catch(XmlPullParserException parser_ex){
            System.out.println(parser_ex.getMessage());
            parser_ex.printStackTrace();
        }
        catch (IOException ioe) {           
            ioe.printStackTrace();            
        }
        String message = errors.size()>0? "" : null;
        for (int i = 0; i < errors.size(); i++) {
            message += (String) errors.get(i)+"\n";
        }
        return message;
    }

    /*
     * (non-Javadoc)
     * @see org.gvsig.remoteClient.wms.WMSProtocolHandler#createGetFeatureInfoRequest(org.gvsig.remoteClient.wms.WMSStatus, int, int)
     */
	protected WMSGetFeatureInfoRequest createGetFeatureInfoRequest(
			WMSStatus status, int x, int y, int featureCount) {
		return new WMSGetFeatureInfoRequest1_1_0(status, this, x, y, featureCount);
	}

	/*
	 * (non-Javadoc)
	 * @see org.gvsig.remoteClient.wms.WMSProtocolHandler#createGetMapRequest(org.gvsig.remoteClient.wms.WMSStatus)
	 */
	protected WMSGetMapRequest createGetMapRequest(WMSStatus status) {
		return new WMSGetMapRequest1_1_0(status, this);
	}

	/*
	 * (non-Javadoc)
	 * @see org.gvsig.remoteClient.wms.WMSProtocolHandler#createGetLegendGraphicRequest(org.gvsig.remoteClient.wms.WMSStatus, java.lang.String)
	 */
	protected WMSGetLegendGraphicRequest createGetLegendGraphicRequest(
			WMSStatus status, String layerName) {
		return new WMSGetLegendGraphicRequest1_1_0(status, this, layerName);
	}

	/* (non-Javadoc)
	 * @see org.gvsig.remoteclient.wms.WMSProtocolHandler#createGetCapabilitiesRequest(org.gvsig.remoteclient.wms.WMSStatus)
	 */	
	protected WMSGetCapabilitiesRequest createGetCapabilitiesRequest(
			WMSStatus status) {
		return new WMSGetCapabilitiesRequest1_1_0(status, this);
	}
    
  }
